# -*- coding: utf-8 -*-
# Copyright 2009-2017, Red Hat, Inc.
# License: GPL-2.0+ <http://spdx.org/licenses/GPL-2.0+>
# See the LICENSE file for more details on Licensing

from __future__ import absolute_import

import os
import os.path
import subprocess
import yaml
import fnmatch
import signal
import json

from libtaskotron import config
from libtaskotron import image_utils
from libtaskotron import os_utils
from libtaskotron import file_utils
from libtaskotron import arch_utils
from libtaskotron.logger import log
from libtaskotron import exceptions as exc
from libtaskotron.directives import resultsdb_directive

try:
    from libtaskotron.ext.disposable import vm
except ImportError, e:
    raise exc.TaskotronImportError(e)


class Executor(object):
    '''Executor executes given task in the format of Ansible playbook. Before
    actual execution, executor decides where the playbook will be executed
    according to the config and command line options: locally, on existing
    machine or in a disposable client.

    :ivar dict arg_data: processed cli arguments with some extra runtime
                         variables
    :ivar task_vm: disposable client, instance of :class:`.TestCloudMachine`
    :ivar bool run_remotely: whether the task is run on a remote machine
    '''

    def __init__(self, arg_data):
        self.arg_data = arg_data
        self.task_vm = None
        self.run_remotely = False

    def _spawn_vm(self, uuid):
        '''Spawn a virtual machine using testcloud.

        :param str uuid: unicode string uuid for the task being executed
        :returns: str ip address of spawned vm
        '''
        log.info('Spawning disposable client')

        env = image_utils.devise_environment(self.arg_data)
        self.task_vm = vm.TestCloudMachine(uuid)
        self.task_vm.prepare(**env)

        self.task_vm.wait_for_port(22)

        log.debug('Disposable client (%s %s) ready', self.task_vm.instancename,
                  self.task_vm.ipaddr)

        return self.task_vm.ipaddr

    def _get_client_ipaddr(self):
        '''Get an IP address of the machine the task is going to be executed
        on.

        :returns: str ip address of the machine or None if the machine is yet
        to be created
        '''
        # when running remotely, run directly over ssh, instead of using
        # libvirt
        persistent = False

        runtask_mode = config.get_config().runtask_mode
        if runtask_mode == config.RuntaskModeName.LOCAL:
            self.run_remotely = False
        elif runtask_mode == config.RuntaskModeName.LIBVIRT:
            self.run_remotely = True
        else:
            assert False, 'This should never occur'

        if self.arg_data['local']:
            log.debug("Forcing local execution (option --local)")
            self.run_remotely = False

        elif self.arg_data['libvirt']:
            log.debug("Forcing remote execution (option --libvirt)")
            self.run_remotely = True
            persistent = False

        elif self.arg_data['ssh']:
            log.debug('Forcing remote execution (option --ssh)')
            self.run_remotely = True
            persistent = True

        log.debug('Execution mode: %s', 'remote' if self.run_remotely
            else 'local')

        ipaddr = '127.0.0.1'
        if self.run_remotely:
            ipaddr = self.arg_data['machine'] if persistent else None

        return ipaddr

    @staticmethod
    def _check_playbook_syntax(playbook):
        '''Run ansible-playbook --syntax-check on a playbook

        :param str playbook: path to an ansible playbook
        :raise TaskotronPlaybookError: when the playbook is not syntactically
            correct
        '''
        try:
            subprocess.check_call(['ansible-playbook', '--syntax-check',
                playbook])
            log.debug('Playbook is syntactically correct: %s', playbook)
        except subprocess.CalledProcessError, e:
            log.error('Syntax check failed for playbook %s: %s', playbook,
                      e)
            raise exc.TaskotronPlaybookError(e)

    def _run_playbook(self, test_playbook, ipaddr, root=True):
        '''Run the ansible-playbook command to execute given playbook
        containing the task.

        :param str test_playbook: name of the playbook, relative to the task
            directory
        :param str ipaddr: IP address of the machine the task will be run on
        :param bool root: whether to run as ``root`` for local execution mode
        :return: stream output of the ansible-playbook command (stdout and
            stderr merged together)
        :rtype: str
        :raise TaskotronPlaybookError: when the playbook is not syntactically
            correct
        '''
        # syntax check
        self._check_playbook_syntax(
            os.path.join(self.arg_data['taskdir'], test_playbook))

        # compute variables
        ansible_dir = os.path.join(config.get_config()._data_dir, 'ansible')
        artifacts_subdir = os.path.join(self.arg_data['artifactsdir'],
                                        test_playbook)

        if os.path.isfile(os.path.join(self.arg_data['taskdir'], 'inventory')):
            sti_inventory = os.path.join(config.get_config().client_taskdir,
                                        'inventory')
        else:
            sti_inventory = "/usr/share/ansible/inventory"

        supported_arches = config.get_config().supported_arches
        supported_binary_arches = [binarch for arch in supported_arches
            for binarch in arch_utils.Arches.binary[arch]]

        minion_repos = config.get_config().minion_repos

        # load taskotron vars from playbook and compute variables
        with open(os.path.join(self.arg_data['taskdir'], test_playbook), 'r') \
            as playbook_file:
            # playbook should contain at least one play as the file passed the
            # syntax check
            # TODO: should we look for 'taskotron_generic_task' in all plays,
            # not just in the first?
            playbook = yaml.load(playbook_file.read())[0]
            if playbook.get('vars',{}).get('taskotron_generic_task', False):
                exec_tasks = 'tasks_generic.yml'
                log.debug('Playbook %s is a generic task', test_playbook)
            else:
                exec_tasks = 'tasks_sti.yml'
                log.debug('Playbook %s is an STI task', test_playbook)
            keepalive = playbook.get('vars', {}).get(
                'taskotron_keepalive_minutes', None)

        # set up variables to load into playbooks.
        # separate them into the ones to forward, and the ones to use
        # internally

        # if you adjust forwarded variables, don't forget to adjust docs
        # as well
        ansible_vars = {
            'artifacts': artifacts_subdir,
            'taskotron_arch': self.arg_data['arch'],
            'taskotron_item': self.arg_data['item'],
            'taskotron_item_type': self.arg_data['type'],
            'taskotron_supported_arches': supported_arches,
            'taskotron_supported_binary_arches': supported_binary_arches,
        }
        varsfile = os.path.join(artifacts_subdir, 'taskotron',
                                'ansible_vars.json')

        ansible_vars_internal = {
            'artifacts_root': self.arg_data['artifactsdir'],
            'client_taskdir': config.get_config().client_taskdir,
            'exec_tasks': exec_tasks,
            'minion_repos': minion_repos,
            'taskdir': self.arg_data['taskdir'],
            'test_playbook': test_playbook,
            'sti_inventory': sti_inventory,
            'varsfile': os.path.basename(varsfile),
        }
        varsfile_internal = os.path.join(artifacts_subdir, 'taskotron',
                                         'ansible_vars_internal.json')

        if keepalive:
            ansible_vars_internal['keepalive_minutes'] = keepalive

        # figure out the ansible-playbook command
        cmd = [
            'ansible-playbook', 'runner.yml',
            '--inventory=%s,' % ipaddr,  # the ending comma is important
            '--extra-vars=@%s' % varsfile,
            '--extra-vars=@%s' % varsfile_internal,
        ]

        if self.run_remotely:
            if self.arg_data['ssh_privkey']:
                cmd.append('--private-key=%s' % self.arg_data['ssh_privkey'])
        else:
            cmd.append('--connection=local')
            ansible_vars_internal['local'] = True
            if root:
                cmd.append('--become')
                ansible_vars_internal['become_root'] = True
            else:
                ansible_vars_internal['become_root'] = False

        if self.arg_data['debug']:
            cmd.append('-vv')

        # store the variables in json files, so that ansible can load them
        file_utils.makedirs(os.path.join(artifacts_subdir, 'taskotron'))
        with open(varsfile, 'w') as vf:
            vars_str = json.dumps(ansible_vars, indent=2, sort_keys=True)
            vf.write(vars_str)
            log.debug('Saved ansible vars file %s with contents:\n%s',
                      varsfile, vars_str)
        with open(varsfile_internal, 'w') as vf:
            vars_str = json.dumps(ansible_vars_internal, indent=2,
                                  sort_keys=True)
            vf.write(vars_str)
            log.debug('Saved internal ansible vars file %s with contents:\n%s',
                      varsfile_internal, vars_str)

        log.debug('Running ansible playbook %s', ' '.join(cmd))
        try:
            # during playbook execution, handle system signals asking us to
            # quit
            signal.signal(signal.SIGINT, self._interrupt_handler)
            signal.signal(signal.SIGTERM, self._interrupt_handler)

            output, _ = os_utils.popen_rt(cmd, cwd=ansible_dir)
            return output
        except subprocess.CalledProcessError as e:
            log.error('ansible-playbook ended with %d return code',
                      e.returncode)
            log.debug(e.output)
            raise exc.TaskotronError(e)
        except exc.TaskotronInterruptError as e:
            log.error('System interrupt %s detected. Pulling logs and '
                      'stopping execution.', e)
            cmd_failsafe = cmd + ['--tags', 'failsafe']
            try:
                os_utils.popen_rt(cmd_failsafe, cwd=ansible_dir)
            except (subprocess.CalledProcessError,
                    exc.TaskotronInterruptError) as e2:
                log.error('Error during failsafe pulling logs, ignoring and '
                    'raising the original error. The current error is: %s:%s',
                    e2.__class__.__name__, e2)
            raise e
        finally:
            # reset signal handling to default behavior
            signal.signal(signal.SIGINT, signal.SIG_DFL)
            signal.signal(signal.SIGTERM, signal.SIG_DFL)

    def _interrupt_handler(self, signum, frame):
        '''Catch system signals (like SIGINT, SIGTERM) and raise them
        as a TasktoronInterruptError'''
        signals_to_names = dict((getattr(signal, n), n) for n in dir(signal)
            if n.startswith('SIG') and '_' not in n )
        signame = signals_to_names.get(signum, 'UNKNOWN')

        log.warn('Received system signal %d (%s). Raising exception.',
            signum, signame)
        raise exc.TaskotronInterruptError(signum, signame)

    def _report_results(self, test_playbook):
        results_file = os.path.join(self.arg_data['artifactsdir'],
            test_playbook, 'taskotron', 'results.yml')
        if os.path.exists(results_file):
            log.info('Reporting results from: %s', results_file)
            rdb = resultsdb_directive.ResultsdbDirective()
            rdb.process(params={"file": results_file}, arg_data=self.arg_data)
        else:
            #FIXME change to exception?
            log.warn("Results file %s doesn't exist" % results_file)

    def execute(self):
        '''Execute all the tasks in the taskdir

        :return: ``True`` if execution finished (that doesn't mean all
                 playbooks finished successfully). ``False`` if the execution
                 was interrupted (e.g. a system signal).
        :rtype: bool
        '''
        test_playbooks = fnmatch.filter(os.listdir(self.arg_data['taskdir']),
                                                   'tests*.yml')
        if not test_playbooks:
            raise exc.TaskotronError('No tests*.yml found in dir %s' %
                                     self.arg_data['taskdir'])

        ipaddr = self._get_client_ipaddr()
        if ipaddr is None:
            ipaddr = self._spawn_vm(self.arg_data['uuid'])

        log.info('Running task on machine: %s', ipaddr)

        interrupted = False
        for test_playbook in test_playbooks:
            log.info('Running testsuite: %s' % test_playbook)
            try:
                self._run_playbook(test_playbook, ipaddr)
                self._report_results(test_playbook)
            except exc.TaskotronInterruptError as e:
                log.error('Caught system interrupt during execution of '
                    'testuite %s: %s. Not executing any other playbooks.',
                    test_playbook, e)
                interrupted = True
                break
            except exc.TaskotronError as e:
                log.error('Error during execution of testsuite %s: %s',
                    test_playbook, e)
            finally:
                if self.task_vm is not None:
                    self.task_vm.teardown()
                log.info('Testsuite execution finished: %s', test_playbook)

        return not interrupted
