# -*- coding: utf-8 -*-
# Copyright 2009-2017, Red Hat, Inc.
# License: GPL-2.0+ <http://spdx.org/licenses/GPL-2.0+>
# See the LICENSE file for more details on Licensing

from __future__ import absolute_import

import re

from libtaskotron import exceptions as exc
from libtaskotron.logger import log

try:
    from libtaskotron.ext.fedora import rpm_utils
except ImportError, e:
    raise exc.TaskotronImportError(e)


def devise_environment(arg_data):
    '''Takes an input item and type, and returns a required run-environment (i.e.
    distro, arch, fedora release, and base-image flavor), based on item and type.

    :param dict formula: parsed formula file (or dict with equivalent structure)
    :param dict arg_data: parsed command-line arguments. item, type and arch
                          are used in this method
    :return: dict containing distro, release, flavor, arch. Each either set, or None
    :raise TaskotronValueError: when environment can't be parsed from the formula
    '''

    env = {'distro': None, 'release': None, 'flavor': None, 'arch': None}

    item = arg_data.get('item', None)
    item_type = arg_data.get('type', None)

    if not env['distro']:
        if item_type == 'koji_build':
            # FIXME: find a way to make this not Fedora-specific
            # For `xchat-2.8.8-21.fc20` disttag is `fc20` for example
            try:
                distro = rpm_utils.get_dist_tag(item)[:2]
                env['distro'] = {'fc': 'fedora'}.get(distro, None)
            except exc.TaskotronValueError:
                env['distro'] = None

        elif item_type == 'koji_tag':
            if re.match(r'^f[0-9]{2}-.*', item):
                env['distro'] = 'fedora'

        if env['distro']:
            log.debug("Environment/distro overriden with ENVVAR from %r:%r to %r",
                      item_type, item, env['distro'])
        else:
            log.debug("Environment/distro can not be inferred from %r:%r. Using default.",
                      item_type, item)

    if not env['release']:
        if item_type == 'koji_build':
            # FIXME: find a way to make this not Fedora-specific
            # Last two characters in rpm's disttag are the Fedora release.
            # For `xchat-2.8.8-21.fc20` disttag is `fc20` for example
            try:
                env['release'] = rpm_utils.get_dist_tag(item)[-2:]
            except exc.TaskotronValueError:
                env['release'] = None

        elif item_type == 'koji_tag':
            if re.match(r'^f[0-9]{2}-.*', item):
                env['release'] = item[1:3]

        if env['release']:
            log.debug("Environment/release inferred from %r:%r to %r",
                      item_type, item, env['release'])
        else:
            log.debug("Environment/release can not be inferred from %r:%r. Using default.",
                      item_type, item)

    if not env['flavor']:
        log.debug("Environment/flavor not specified. Using default")

    if not env['arch']:
        arch = arg_data.get('arch')
        if not arch or arch == 'noarch':
            log.warn("Environment/arch can not be inferred from %r. Using default.", arch)
        else:
            env['arch'] = arch
            log.debug("Environment/arch inferred from %r to %r", arch, env['arch'])

    return env
