import ast
from libtaskotron.ext.fedora.koji_utils import KojiClient
from libtaskotron.ext.fedora import rpm_utils
import libtaskotron.exceptions as exc

class KojiDirective(object):
    # FIXME: don't duplicate code with koji_directive.py

    def __init__(self, koji_session=None):
        super(KojiDirective, self).__init__()
        if koji_session is None:
            self.koji = KojiClient()
        else:
            self.koji = koji_session

    def process(self, mod):
        # process params
        valid_actions = ['download', 'download_tag', 'download_latest_stable']
        action = mod.params['action']
        if action not in valid_actions:
            raise exc.TaskotronDirectiveError('%s is not a valid action for koji '
                                                'directive' % action)

        if 'arch' not in mod.params:
            detected_args = ', '.join(mod.params.keys())
            raise exc.TaskotronDirectiveError(
                "The koji directive requires 'arch' as an argument. Detected "
                "arguments: %s" % detected_args)

        # this is supposedly safe enough to use on raw input but should be double checked
        # http://stackoverflow.com/questions/1894269/convert-string-representation-of-list-to-list-in-python
        arches = ast.literal_eval(mod.params['arch'])
        if not isinstance(arches, list):
            raise exc.TaskotronError("arches must be a list")

        if arches and ('all' not in arches) and ('noarch' not in arches):
            arches.append('noarch')

        arch_exclude_string = mod.params.get('arch_exclude', None)
        if arch_exclude_string is None:
            arch_exclude = []
        else:
            arch_exclude = ast.literal_eval(mod.params['arch_exclude'])

        debuginfo = mod.params.get('debuginfo', False)
        src = mod.params.get('src', False)
        build_log = mod.params.get('build_log', False)

        if not isinstance(arch_exclude, list):
            print("arch_exclude: {}".format(type(arch_exclude)))
            raise Exception("arch_exclude must be a list")
        # download files
        output_data = {}

        if action == 'download':
            if 'koji_build' not in mod.params:
                detected_args = ', '.join(mod.params.keys())
                raise exc.TaskotronDirectiveError(
                    "The koji directive requires 'koji_build' for the 'download' "
                    "action. Detected arguments: %s" % detected_args)

            nvr = rpm_utils.rpmformat(mod.params['koji_build'], 'nvr')
            output_data['downloaded_rpms'] = self.koji.get_nvr_rpms(
                nvr, mod.params['target_dir'], arches=arches, arch_exclude=arch_exclude,
                debuginfo=debuginfo, src=src)

        elif action == 'download_tag':
            if 'koji_tag' not in mod.params:
                detected_args = ', '.join(mod.params.keys())
                raise exc.TaskotronDirectiveError(
                    "The koji directive requires 'koji_tag' for the 'download_tag' "
                    "action. Detected arguments: %s" % detected_args)

            koji_tag = mod.params['koji_tag']

            output_data['downloaded_rpms'] = self.koji.get_tagged_rpms(
                koji_tag, mod.params['target_dir'], arches=arches, arch_exclude=arch_exclude,
                debuginfo=debuginfo, src=src)

        elif action == 'download_latest_stable':
            if 'koji_build' not in mod.params:
                detected_args = ', '.join(mod.params.keys())
                raise exc.TaskotronDirectiveError(
                    "The koji directive requires 'koji_build' for the 'download_latest_stable' "
                    "action. Detected arguments: %s" % detected_args)

            name = rpm_utils.rpmformat(mod.params['koji_build'], 'n')
            disttag = rpm_utils.get_dist_tag(mod.params['koji_build'])
            # we need to do 'fc22' -> 'f22' conversion
            tag = disttag.replace('c', '')

            # first we need to check updates tag and if that fails, the latest
            # stable nvr is in the base repo
            tags = ['%s-updates' % tag, tag]
            nvr = self.koji.latest_by_tag(tags, name)

            output_data['downloaded_rpms'] = self.koji.get_nvr_rpms(
                nvr, mod.params['target_dir'], arch_exclude=arch_exclude,
                arches=arches, debuginfo=debuginfo, src=src)

        # download build.log if requested
        if build_log:
            if action in ('download', 'download_latest_stable'):
                ret_log = self.koji.get_build_log(
                        nvr, mod.params['target_dir'], arches=arches, arch_exclude=arch_exclude)
                output_data['downloaded_logs'] = ret_log['ok']
                output_data['log_errors'] = ret_log['error']
            else:
                #log.warn("Downloading build logs is not supported for action '%s', ignoring.",
                #         action)
                print("Downloading build logs is not supported for action '%s', ignoring." % action)

        return output_data